package org.dimx;

import cleoni.adv.*;

import java.util.Vector;

/**
 * An implementation of a custom command processor for the DimensioneX Game Engine.
 * The developer can use or subclass as needed.
 * 
 * The custom processor should be defined as a WORLD attribute within the 
 * MyWorld.DXW file.
 * 
 * Example:
 * <pre>
 * WORLD
 *     ...
 *     custCmdProc   org.dimx.CustCmdProcessor
 *     ...
 *     OTHER_TAGS
 * 
 *     END_OTHER_TAGS
 * 
 * END_WORLD
 * </pre>
 * @author Carl Nagle
 */
public class CustCmdProcessor implements ISayProcessor, IWorldInitProcessor{

    /**
     * Display ALL player SAY messages and chat in the Message panel.
     * 
     * If the player selects another player for the SAY then that player will 
     * receive the message but the selected player will NOT receiver player focus.
     * 
     * If the player sends a generic message with [ENTER], or selects a non-player 
     * object then normal SAY processing occurs in addition to displaying the text 
     * in the player's message panel.
     * 
     * The API is subject to change as the parent Interface becomes fully developed.
     * @param server
     * @param world
     * @param cmd
     * @param arg1
     * @param arg2
     * @param player
     * @return NULL to continue with normal say processing; Vector(cmd,arg1,arg2) to override.
     */
    public Vector processSay(multiplayer server, World world, String cmd, String arg1, String arg2, Player player) {
        try{
            //cmd = "say";
            player.getClient().receive(new Message(player, player.getId(), arg1));
            AdvObject target = world.getObject(arg2);
            player.getClient().cmdNewmsg(); // clear the txtBox on next refresh
            if(target != null){
                if(target.isaCharacter()){
                    //talk to a player without changing focus
                    target.hear(player, arg1);
                    return clearSayVector();  //bypass all other processing                
                }
            }            
        }catch(DimxException x){}
        return null; // perform normal SAY processing
    }

    /**
     * Sample custom processor simply logs a debug message and returns 'true' to 
     * allow normal world.init to occur.
     * This API is subject to change until it is fully developed.
     * @param server
     * @param loader
     * @param world
     * @param systemDir
     * @return boolean true to allow normal world.init processing to continue;
     *         false to bypass normal world.init processing.
     */    
    public boolean processPreWorldInit(multiplayer server, WorldLoader loader, World world, String systemDir) {
        world.logger.debug("Custom pre-init processing complete.");
        return true;
    }

    /**
     * Sample custom processor simply logs debug messages and returns 'true' to 
     * allow normal world.init to finalize.
     * This API is subject to change until it is fully developed.
     * @param server
     * @param loader
     * @param world
     * @param systemDir
     * @return boolean true to allow remaining init processing to continue;
     *         false to bypass remaining init processing.
     */
    public boolean processPostWorldInit(multiplayer server, WorldLoader loader, World world, String systemDir) {
        Dict _objs = world.getObjects();
        world.logger.debug("Custom post-init processing of "+ _objs.size()+ " objects commencing.");
        // Do stuff in here.
        world.logger.debug("Custom post-init processing complete.");
        return true;
    }

    // used to "do nothing". Essentially, we've already done all we want to do
    private Vector clearSayVector() {
        Vector v = new Vector(3); /* prepares the result */
        v.addElement("");
        v.addElement("");
        v.addElement("");
        return v;                
    }
    
}
